/*
 * JOxy.java
 *
 * Created on 20 agosto 2013
 * Revised on 20 agosto 2014
 */
package beans;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.image.RenderedImage;
import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.text.NumberFormat;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import javax.swing.JComponent;

/**
 *
 * @author Gianfranco Oddenino
 */
public class JOxy extends JComponent implements Serializable {

    private Image image;
    private Point penPos;
    private int width, height;
    private String titleX = "X", titleY = "Y";
    private boolean gridX = true, gridY = true;
    private double centerX = 0.0, centerY = 0.0, scaleX = 32.0, scaleY = 32.0;
    private Graphics graphics;
    private NumberFormat nf;
    private FontMetrics fm;
    private int fh, fa;
    private double x1, x2, dx, y1, y2, dy;
    private int nx, ny;

    public JOxy() {
        setPreferredSize(new Dimension(200, 200));
        setMinimumSize(new Dimension(200, 200));
        width = getWidth();
        height = getHeight();
        EventiMouse em = new EventiMouse();
        addMouseListener(em);
        addMouseMotionListener(em);
        addMouseWheelListener(em);
    }

    private class EventiMouse extends MouseAdapter {

        private double centerXold, centerYold;

        @Override
        public void mousePressed(MouseEvent evt) {
            centerXold = centerX + evt.getX() / scaleX;
            centerYold = centerY - evt.getY() / scaleY;
        }

        @Override
        public void mouseDragged(MouseEvent evt) {
            centerX = centerXold - evt.getX() / scaleX;
            centerY = centerYold + evt.getY() / scaleY;
            clear();
        }

        @Override
        public void mouseWheelMoved(MouseWheelEvent evt) {
            if (evt.getWheelRotation() > 0) {
                double x = xw(evt.getX()), y = yw(evt.getY());
                centerX -= (x - centerX) / 9.0;
                centerY -= (y - centerY) / 9.0;
                scaleX *= 9.0 / 10.0;
                scaleY *= 9.0 / 10.0;
            } else {
                double x = xw(evt.getX()), y = yw(evt.getY());
                centerX += (x - centerX) / 10.0;
                centerY += (y - centerY) / 10.0;
                scaleX *= 10.0 / 9.0;
                scaleY *= 10.0 / 9.0;
            }
            clear();
        }

    }

    /**
     * @return a nice number for ticks interval
     */
    private double getNiceNumber(double x) {
        double e, f, n;
        e = Math.floor(Math.log10(x));
        f = x / Math.pow(10., e);
        if (f <= 1.0)
            n = 1.0;
        else if (f <= 2.0)
            n = 2.0;
        else if (f <= 5.0)
            n = 5.0;
        else
            n = 10.0;
        return n * Math.pow(10.0, e);
    }

    /**
     * @clear graphic plot
     */
    public void clear() {
        /*
         * Evita che venga effettuato il disegno degli assi in fase di esecuzione,
         * in cui (sembra) vengono assegnati i valori alle propriet prima della
         * creazione dell'oggetto JOxy, con un conseguente null-pointer exception
         * sull'oggetto image.
         */
        if (image == null)
            return;

        nf = NumberFormat.getNumberInstance();
        fm = graphics.getFontMetrics();
        fh = fm.getHeight();
        fa = fm.getAscent();

        dx = getNiceNumber(50 / scaleX);
        x1 = Math.ceil(getMinX() / dx) * dx;
        x2 = Math.floor(getMaxX() / dx) * dx;
        nx = (int) Math.round((x2 - x1) / dx);

        dy = getNiceNumber(50 / scaleY);
        y1 = Math.ceil(getMinY() / dy) * dy;
        y2 = Math.floor(getMaxY() / dy) * dy;
        ny = (int) Math.round((y2 - y1) / dy);

        drawBackground();   // draw background
        if (gridX)
            drawGridX(); // draw x grid
        if (gridY)
            drawGridY(); // draw y grid
        drawAxisX();    // draw x axis
        drawTitleX();   // draw x axis title
        drawAxisY();    // draw y axis
        drawTitleY();   // draw y axis title
        penPos = new Point(xg(0.0), yg(0.0));
        repaint();
    }

    /**
     * @draw Background
     */
    private void drawBackground() {
        graphics.setColor(new Color(0xFFFFFF));
        graphics.fillRect(0, 0, width, height);
    }

    /**
     * @draw GridX
     */
    private void drawGridX() {
        graphics.setColor(new Color(0xF0F0F0));
        for (int i = 0; i <= nx; i++) {
            int xi = xg(x1 + i * dx);
            graphics.drawLine(xi, 0, xi, height);
        }
    }

    /**
     * @draw GridY
     */
    private void drawGridY() {
        graphics.setColor(new Color(0xF0F0F0));
        for (int i = 0; i <= ny; i++) {
            int yi = yg(y1 + i * dy);
            graphics.drawLine(0, yi, width, yi);
        }
    }

    /**
     * @draw AxisX
     */
    private void drawAxisX() {
        int yi = yg(0.0);
        graphics.setColor(new Color(0x808080));
        graphics.drawLine(0, yi, width, yi);
        nf.setMaximumFractionDigits(-(int) Math.floor(Math.log10(dx)));
        for (int i = 0; i <= nx; i++) {
            int xi = xg(x1 + i * dx);
            graphics.setColor(new Color(0x808080));
            graphics.drawLine(xi, yi - fh / 4, xi, yi + fh / 4);
            graphics.setColor(new Color(0x000000));
            double xVal = x1 + i * dx;
            String num = nf.format(xVal);
            graphics.drawString(num, xi - fm.stringWidth(num) / 2, yi + fh * 9 / 8);
        }
    }

    /**
     * @draw AxisY
     */
    private void drawAxisY() {
        int xi = xg(0.0);
        graphics.setColor(new Color(0x808080));
        graphics.drawLine(xi, 0, xi, height);
        nf.setMaximumFractionDigits(-(int) Math.floor(Math.log10(dy)));
        for (int i = 0; i <= ny; i++) {
            int yi = yg(y1 + i * dy);
            graphics.setColor(new Color(0x808080));
            graphics.drawLine(xi - fh / 4, yi, xi + fh / 4, yi);
            graphics.setColor(new Color(0x000000));
            double yVal = y1 + i * dy;
            String num = nf.format(yVal);
            graphics.drawString(num, xi - fm.stringWidth(num) - fh / 2, yi + fa / 2 - 1);
        }
    }

    /**
     * @draw TitleX
     */
    private void drawTitleX() {
        graphics.setColor(new Color(0x000000));
        graphics.drawString(titleX, width - fh / 4 - fm.stringWidth(titleX), yg(0.0) - fh / 2);
    }

    /**
     * @draw TitleY
     */
    private void drawTitleY() {
        graphics.setColor(new Color(0x000000));
        graphics.drawString(titleY, xg(0.0) + fh / 2, fh);
    }

    @Override
    protected void paintComponent(Graphics g) {
        // super.paintComponent(g);
        if (image == null || width != getWidth() || height != getHeight()) {
            width = getWidth();
            height = getHeight();
            image = createImage(width, height);
            graphics = image.getGraphics();
            clear();
        }
        g.drawImage(image, 0, 0, null);
    }

    public int xg(double x) {
        double xr = width / 2 + (x - centerX) * scaleX;
        if (xr < Integer.MIN_VALUE)
            return Integer.MIN_VALUE;
        else if (xr > Integer.MAX_VALUE)
            return Integer.MAX_VALUE;
        else
            return (int) Math.round(xr);
    }

    public int yg(double y) {
        double yr = height / 2 - (y - centerY) * scaleY;
        if (yr < Integer.MIN_VALUE)
            return Integer.MIN_VALUE;
        else if (yr > Integer.MAX_VALUE)
            return Integer.MAX_VALUE;
        else
            return (int) Math.round(yr);
    }

    public double xw(int x) {
        return centerX + ((double) x - width / 2) / scaleX;
    }

    public double yw(int y) {
        return centerY - ((double) y - height / 2) / scaleY;
    }

    public void moveTo(double x, double y) {
        penPos.x = xg(x);
        penPos.y = yg(y);
        //  repaint();
    }

    public void lineTo(double x, double y) {
        int xr = xg(x);
        int yr = yg(y);
        graphics.drawLine(penPos.x, penPos.y, xr, yr);
        penPos.x = xr;
        penPos.y = yr;
        repaint();
    }

    public void line(double x1, double y1, double x2, double y2) {
        int x1r = xg(x1);
        int y1r = yg(y1);
        int x2r = xg(x2);
        int y2r = yg(y2);
        graphics.drawLine(x1r, y1r, x2r, y2r);
        penPos.x = x2r;
        penPos.y = y2r;
        repaint();
    }

    public void pixel(double x, double y) {
        int xr = xg(x);
        int yr = yg(y);
        graphics.drawLine(xr, yr, xr, yr);
        penPos.x = xr;
        penPos.y = yr;
        repaint();
    }

    public void rectangle(double x1, double y1, double x2, double y2) {
        graphics.drawRect(xg(x1), yg(y2), (int) Math.round((x2 - x1) * scaleX), (int) Math.round((y2 - y1) * scaleY));
        repaint();
    }

    public void ellipse(double x, double y, double a, double b) {
        graphics.drawOval(xg(x - a), yg(y + b), (int) Math.round(2 * a * scaleX), (int) Math.round(2 * b * scaleY));
        repaint();
    }

    public void circle(double x, double y, double r) {
        graphics.drawOval(xg(x - r), yg(y + r), (int) Math.round(2 * r * scaleX), (int) Math.round(2 * r * scaleY));
        repaint();
    }

    public Color getColor() {
        return graphics.getColor();
    }

    public void setColor(Color c) {
        graphics.setColor(c);
    }

    /**
     * @return the penPos
     */
    public Point getPenPos() {
        return penPos;
    }

    /**
     * @return the image
     */
    public Image getImage() {
        return image;
    }

    /**
     * @save the image to file Oxy.png
     */
    public void saveToFile() {
        try {
            File file = new File("Oxy.png");
            ImageIO.write((RenderedImage) image, "png", file);
        } catch (IOException ex) {
            Logger.getLogger(JOxy.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * @return the titleX
     */
    public String getTitleX() {
        return titleX;
    }

    /**
     * @param titleX the titleX to set
     */
    public void setTitleX(String titleX) {
        this.titleX = titleX;
        clear();
    }

    /**
     * @return the titleY
     */
    public String getTitleY() {
        return titleY;
    }

    /**
     * @param titleY the titleY to set
     */
    public void setTitleY(String titleY) {
        this.titleY = titleY;
        clear();
    }

    /**
     * @return the gridX
     */
    public boolean isGridX() {
        return gridX;
    }

    /**
     * @param gridX the gridX to set
     */
    public void setGridX(boolean gridX) {
        this.gridX = gridX;
        clear();
    }

    /**
     * @return the gridY
     */
    public boolean isGridY() {
        return gridY;
    }

    /**
     * @param gridY the gridY to set
     */
    public void setGridY(boolean gridY) {
        this.gridY = gridY;
        clear();
    }

    /**
     * @return the maxX
     */
    public double getMaxX() {
        return width / (2 * scaleX) + centerX;
    }

    /**
     * @return the maxY
     */
    public double getMaxY() {
        return height / (2 * scaleY) + centerY;
    }

    /**
     * @return the minX
     */
    public double getMinX() {
        return -width / (2 * scaleX) + centerX;
    }

    /**
     * @return the minY
     */
    public double getMinY() {
        return -height / (2 * scaleY) + centerY;
    }

    /**
     * @return the centerX
     */
    public double getCenterX() {
        return centerX;
    }

    /**
     * @param centerX the centerX to set
     */
    public void setCenterX(double centerX) {
        this.centerX = centerX;
        clear();
    }

    /**
     * @return the centerY
     */
    public double getCenterY() {
        return centerY;
    }

    /**
     * @param centerY the centerY to set
     */
    public void setCenterY(double centerY) {
        this.centerY = centerY;
        clear();
    }

    /**
     * @return the scaleX
     */
    public double getScaleX() {
        return scaleX;
    }

    /**
     * @param scaleX the scaleX to set
     */
    public void setScaleX(double scaleX) {
        this.scaleX = scaleX;
        clear();
    }

    /**
     * @return the scaleY
     */
    public double getScaleY() {
        return scaleY;
    }

    /**
     * @param scaleY the scaleY to set
     */
    public void setScaleY(double scaleY) {
        this.scaleY = scaleY;
        clear();
    }

}
