/*
 * JPlot.java
 *
 * Created on 1 agosto 2013
 * Revised on 20 agosto 2014
 */
package beans;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.geom.AffineTransform;
import java.awt.image.RenderedImage;
import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.text.NumberFormat;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import javax.swing.JComponent;

/**
 *
 * @author Gianfranco Oddenino
 */
public class JPlot extends JComponent implements Serializable {

    private Image image;
    private Point penPos;
    private int width, height;
    private String title1 = "Main title", title2 = "Plot description", titleX = "X", titleY = "Y";
    private boolean gridX = true, gridY = true;
    private double maxX = 1.0, maxY = 1.0, minX = 0.0, minY = 0.0;
    private Graphics graphics;
    private Rectangle viewport;
    private Rectangle.Double window;
    private NumberFormat nf;
    private FontMetrics fm;
    private int fh, fa;
    private double x1, x2, dx, y1, y2, dy, expX, expY, scX, scY;
    private int nx, ny;

    public JPlot() {
        setPreferredSize(new Dimension(200, 200));
        setMinimumSize(new Dimension(200, 200));
        width = getWidth();
        height = getHeight();
        EventiMouse em = new EventiMouse();
        addMouseListener(em);
        addMouseMotionListener(em);
        addMouseWheelListener(em);
    }

    private class EventiMouse extends MouseAdapter {

        private double maxXold, maxYold, minXold, minYold;

        @Override
        public void mousePressed(MouseEvent evt) {
            minXold = minX + evt.getX() * window.width / viewport.width;
            minYold = minY - evt.getY() * window.height / viewport.height;
            maxXold = maxX + evt.getX() * window.width / viewport.width;
            maxYold = maxY - evt.getY() * window.height / viewport.height;
        }

        @Override
        public void mouseDragged(MouseEvent evt) {
            minX = minXold - evt.getX() * window.width / viewport.width;
            minY = minYold + evt.getY() * window.height / viewport.height;
            maxX = maxXold - evt.getX() * window.width / viewport.width;
            maxY = maxYold + evt.getY() * window.height / viewport.height;
            clear();
        }

        @Override
        public void mouseWheelMoved(MouseWheelEvent evt) {
            if (evt.getWheelRotation() > 0) {
                double x = xw(evt.getX()), y = yw(evt.getY());
                minX = x - (x - minX) * 10.0 / 9.0;
                maxX = x + (maxX - x) * 10.0 / 9.0;
                minY = y - (y - minY) * 10.0 / 9.0;
                maxY = y + (maxY - y) * 10.0 / 9.0;
            } else {
                double x = xw(evt.getX()), y = yw(evt.getY());
                minX = x - (x - minX) * 9.0 / 10.0;
                maxX = x + (maxX - x) * 9.0 / 10.0;
                minY = y - (y - minY) * 9.0 / 10.0;
                maxY = y + (maxY - y) * 9.0 / 10.0;
            }
            clear();
        }

    }

    /**
     * @return a nice number for ticks interval
     */
    private double getNiceNumber(double x) {
        double e, f, n;
        e = Math.floor(Math.log10(x));
        f = x / Math.pow(10., e);
        if (f <= 1.0)
            n = 1.0;
        else if (f <= 2.0)
            n = 2.0;
        else if (f <= 5.0)
            n = 5.0;
        else
            n = 10.0;
        return n * Math.pow(10.0, e);
    }

    /**
     * @clear graphic plot
     */
    public void clear() {
        /*
         * Evita che venga effettuato il disegno degli assi in fase di esecuzione,
         * in cui (sembra) vengono assegnati i valori alle propriet prima della
         * creazione dell'oggetto JPlot, con un conseguente null-pointer exception
         * sull'oggetto image.
         */
        if (image == null)
            return;

        nf = NumberFormat.getNumberInstance();
        fm = graphics.getFontMetrics();
        fh = fm.getHeight();
        fa = fm.getAscent();
        viewport = new Rectangle(fh * 9 / 2, fh * 3, width - fh * 6, height - fh * 6);
        window = new Rectangle.Double(minX, minY, maxX - minX, maxY - minY);

        double xMag = Math.max(Math.abs(minX), Math.abs(maxX));
        if (xMag < 0.1 || xMag >= 1000.0) {
            expX = Math.floor(Math.log10(xMag));
            scX = Math.pow(10.0, expX);
        } else {
            expX = 0.0;
            scX = 1.0;
        }
        dx = getNiceNumber(50 * window.width / viewport.width);
        x1 = Math.ceil(window.x / dx) * dx;
        x2 = Math.floor((window.x + window.width) / dx) * dx;
        nx = (int) Math.round((x2 - x1) / dx);

        double yMag = Math.max(Math.abs(minY), Math.abs(maxY));
        if (yMag < 0.1 || yMag >= 1000.0) {
            expY = Math.floor(Math.log10(yMag));
            scY = Math.pow(10.0, expY);
        } else {
            expY = 0.0;
            scY = 1.0;
        }
        dy = getNiceNumber(50 * window.height / viewport.height);
        y1 = Math.ceil(window.y / dy) * dy;
        y2 = Math.floor((window.y + window.height) / dy) * dy;
        ny = (int) Math.round((y2 - y1) / dy);

        graphics.setClip(null);
        drawBackground();   // draw background
        if (gridX)
            drawGridX(); // draw x grid
        if (gridY)
            drawGridY(); // draw y grid
        drawFrame();    // draw frame
        drawAxisX();    // draw x axis
        drawTitleX();   // draw x axis title
        drawAxisY();    // draw y axis
        drawTitleY();   // draw y axis title
        drawTitle1();   // draw first title
        drawTitle2();   // draw second title
        graphics.setClip(viewport.x, viewport.y, viewport.width + 1, viewport.height + 1);
        penPos = new Point(viewport.x, viewport.y + viewport.height);
        repaint();
    }

    /**
     * @draw Background
     */
    private void drawBackground() {
        graphics.setColor(new Color(0xF0F0F0));
        graphics.fillRect(0, 0, width, height);
        graphics.setColor(new Color(0xFFFFFF));
        graphics.fillRect(viewport.x, viewport.y, viewport.width + 1, viewport.height + 1);
    }

    /**
     * @draw GridX
     */
    private void drawGridX() {
        graphics.setColor(new Color(0xF0F0F0));
        for (int i = 0; i <= nx; i++) {
            int xi = xg(x1 + i * dx);
            graphics.drawLine(xi, viewport.y, xi, viewport.y + viewport.height);
        }
    }

    /**
     * @draw GridY
     */
    private void drawGridY() {
        graphics.setColor(new Color(0xF0F0F0));
        for (int i = 0; i <= ny; i++) {
            int yi = yg(y1 + i * dy);
            graphics.drawLine(viewport.x, yi, viewport.x + viewport.width, yi);
        }
    }

    /**
     * @draw Frame
     */
    private void drawFrame() {
        graphics.setColor(new Color(0xC0C0C0));
        graphics.drawRect(viewport.x, viewport.y, viewport.width, viewport.height);
    }

    /**
     * @draw AxisX
     */
    private void drawAxisX() {
        graphics.setColor(new Color(0x808080));
        graphics.drawLine(viewport.x, viewport.y + viewport.height, viewport.x + viewport.width, viewport.y + viewport.height);
        nf.setMaximumFractionDigits(-(int) Math.floor(Math.log10(dx / scX)));
        for (int i = 0; i <= nx; i++) {
            int xi = xg(x1 + i * dx);
            graphics.setColor(new Color(0x808080));
            graphics.drawLine(xi, viewport.y + viewport.height, xi, viewport.y + viewport.height + fh / 4);
            graphics.setColor(new Color(0x000000));
            double xVal = x1 + i * dx;
            String num = nf.format(xVal / scX);
            graphics.drawString(num, xi - fm.stringWidth(num) / 2, viewport.y + viewport.height + fh * 9 / 8);
        }
    }

    /**
     * @draw AxisY
     */
    private void drawAxisY() {
        graphics.setColor(new Color(0x808080));
        graphics.drawLine(viewport.x, viewport.y, viewport.x, viewport.y + viewport.height);
        nf.setMaximumFractionDigits(-(int) Math.floor(Math.log10(dy / scY)));
        for (int i = 0; i <= ny; i++) {
            int yi = yg(y1 + i * dy);
            graphics.setColor(new Color(0x808080));
            graphics.drawLine(viewport.x, yi, viewport.x - fh / 4, yi);
            graphics.setColor(new Color(0x000000));
            double yVal = y1 + i * dy;
            String num = nf.format(yVal / scY);
            graphics.drawString(num, viewport.x - fm.stringWidth(num) - fh / 2, yi + fa / 2 - 1);
        }
    }

    /**
     * @draw Title1
     */
    private void drawTitle1() {
        graphics.setColor(new Color(0x000000));
        graphics.setFont(graphics.getFont().deriveFont(Font.BOLD));
        graphics.drawString(title1, width / 2 - fm.stringWidth(title1) / 2, fh * 5 / 4);
    }

    /**
     * @draw Title2
     */
    private void drawTitle2() {
        graphics.setColor(new Color(0x000000));
        graphics.setFont(graphics.getFont().deriveFont(Font.PLAIN));
        graphics.drawString(title2, width / 2 - fm.stringWidth(title2) / 2, fh * 9 / 4);
    }

    /**
     * @draw TitleX
     */
    private void drawTitleX() {
        graphics.setColor(new Color(0x000000));
        String str = titleX;
        if (expX != 0.0)
            str += "     10^" + (int) expX;
        graphics.drawString(str, viewport.x + viewport.width / 2 - fm.stringWidth(str) / 2, height - fh / 2);
    }

    /**
     * @draw TitleY
     */
    private void drawTitleY() {
        graphics.setColor(new Color(0x000000));
        String str = titleY;
        if (expY != 0.0)
            str += "     10^" + (int) expY;
        double[] m = {0.0, -1.0, 1.0, 0.0,
            fh, viewport.y + viewport.height / 2 + fm.stringWidth(str) / 2};
        AffineTransform saveAT = ((Graphics2D) graphics).getTransform();
        ((Graphics2D) graphics).setTransform(new AffineTransform(m));
        ((Graphics2D) graphics).drawString(str, 0, 0);
        ((Graphics2D) graphics).setTransform(saveAT);
    }

    @Override
    protected void paintComponent(Graphics g) {
        // super.paintComponent(g);
        if (image == null || width != getWidth() || height != getHeight()) {
            width = getWidth();
            height = getHeight();
            image = createImage(width, height);
            graphics = image.getGraphics();
            clear();
        }
        g.drawImage(image, 0, 0, null);
    }

    public int xg(double x) {
        double xr = viewport.x + (x - window.x) * viewport.width / window.width;
        if (xr < Integer.MIN_VALUE)
            return Integer.MIN_VALUE;
        else if (xr > Integer.MAX_VALUE)
            return Integer.MAX_VALUE;
        else
            return (int) Math.round(xr);
    }

    public int yg(double y) {
        double yr = viewport.y + viewport.height - (y - window.y) * viewport.height / window.height;
        if (yr < Integer.MIN_VALUE)
            return Integer.MIN_VALUE;
        else if (yr > Integer.MAX_VALUE)
            return Integer.MAX_VALUE;
        else
            return (int) Math.round(yr);
    }

    public double xw(int x) {
        return window.x + (x - viewport.x) * window.width / viewport.width;
    }

    public double yw(int y) {
        return window.y + (viewport.y + viewport.height - y) * window.height / viewport.height;
    }

    public void moveTo(double x, double y) {
        penPos.x = xg(x);
        penPos.y = yg(y);
        //  repaint();
    }

    public void lineTo(double x, double y) {
        int xr = xg(x);
        int yr = yg(y);
        graphics.drawLine(penPos.x, penPos.y, xr, yr);
        penPos.x = xr;
        penPos.y = yr;
        repaint();
    }

    public void line(double x1, double y1, double x2, double y2) {
        int x1r = xg(x1);
        int y1r = yg(y1);
        int x2r = xg(x2);
        int y2r = yg(y2);
        graphics.drawLine(x1r, y1r, x2r, y2r);
        penPos.x = x2r;
        penPos.y = y2r;
        repaint();
    }

    public void pixel(double x, double y) {
        int xr = xg(x);
        int yr = yg(y);
        graphics.drawLine(xr, yr, xr, yr);
        penPos.x = xr;
        penPos.y = yr;
        repaint();
    }

    public Color getColor() {
        return graphics.getColor();
    }

    public void setColor(Color c) {
        graphics.setColor(c);
    }

    /**
     * @return the penPos
     */
    public Point getPenPos() {
        return penPos;
    }

    /**
     * @return the image
     */
    public Image getImage() {
        return image;
    }

    /**
     * @save the image to file Plot.png
     */
    public void saveToFile() {
        try {
            File file = new File("Plot.png");
            ImageIO.write((RenderedImage) image, "png", file);
        } catch (IOException ex) {
            Logger.getLogger(JPlot.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * @return the title1
     */
    public String getTitle1() {
        return title1;
    }

    /**
     * @param title1 the title1 to set
     */
    public void setTitle1(String title1) {
        this.title1 = title1;
        clear();
    }

    /**
     * @return the title2
     */
    public String getTitle2() {
        return title2;
    }

    /**
     * @param title2 the title2 to set
     */
    public void setTitle2(String title2) {
        this.title2 = title2;
        clear();
    }

    /**
     * @return the titleX
     */
    public String getTitleX() {
        return titleX;
    }

    /**
     * @param titleX the titleX to set
     */
    public void setTitleX(String titleX) {
        this.titleX = titleX;
        clear();
    }

    /**
     * @return the titleY
     */
    public String getTitleY() {
        return titleY;
    }

    /**
     * @param titleY the titleY to set
     */
    public void setTitleY(String titleY) {
        this.titleY = titleY;
        clear();
    }

    /**
     * @return the gridX
     */
    public boolean isGridX() {
        return gridX;
    }

    /**
     * @param gridX the gridX to set
     */
    public void setGridX(boolean gridX) {
        this.gridX = gridX;
        clear();
    }

    /**
     * @return the gridY
     */
    public boolean isGridY() {
        return gridY;
    }

    /**
     * @param gridY the gridY to set
     */
    public void setGridY(boolean gridY) {
        this.gridY = gridY;
        clear();
    }

    /**
     * @return the maxX
     */
    public double getMaxX() {
        return maxX;
    }

    /**
     * @param maxX the maxX to set
     */
    public void setMaxX(double maxX) {
        this.maxX = maxX;
        clear();
    }

    /**
     * @return the maxY
     */
    public double getMaxY() {
        return maxY;
    }

    /**
     * @param maxY the maxY to set
     */
    public void setMaxY(double maxY) {
        this.maxY = maxY;
        clear();
    }

    /**
     * @return the minX
     */
    public double getMinX() {
        return minX;
    }

    /**
     * @param minX the minX to set
     */
    public void setMinX(double minX) {
        this.minX = minX;
        clear();
    }

    /**
     * @return the minY
     */
    public double getMinY() {
        return minY;
    }

    /**
     * @param minY the minY to set
     */
    public void setMinY(double minY) {
        this.minY = minY;
        clear();
    }

}
